@@ -18,5 +18,5 @@ def retry_send_messages():
     Celery periodic task for retry send failed messages.
     """
     core._send_pending_messages()
-    core._mark_discarted_messages()
+    core._mark_discarded_messages()
     core._retry_send_messages()
